package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module GeometrySource

Geometric primitives useful for layout.

This module provides foundational types for representing dimensions, positions, and constraints used throughout the layout engine.

Axis types

Sourcemodule Absolute_axis : sig ... end

Absolute horizontal and vertical axis representation.

Sourcemodule Abstract_axis : sig ... end

CSS abstract axis types and operations.

Geometric types

Sourcemodule Size : sig ... end

2-dimensional size representation.

Sourcemodule Point : sig ... end

2-dimensional point representation.

Sourcemodule Line : sig ... end

An abstract line with start and end positions.

Sourcemodule Rect : sig ... end

Axis-aligned rectangles representing margins, padding, borders, or insets.

Sourcemodule Min_max : sig ... end

Generic container for minimum and maximum constraint values.

Sourcemodule In_both_abs_axis : sig ... end

Container that holds an item in each absolute axis.

Type aliases

Sourcetype absolute_axis = Absolute_axis.t

Alias for Absolute_axis.t.

Sourcetype abstract_axis = Abstract_axis.t

Alias for Abstract_axis.t.

Sourcetype 'a rect = 'a Rect.t

Alias for Rect.t.

Sourcetype 'a size = 'a Size.t

Alias for Size.t.

Sourcetype 'a point = 'a Point.t

Alias for Point.t.

Sourcetype 'a line = 'a Line.t

Alias for Line.t.

Sourcetype ('min, 'max) min_max = ('min, 'max) Min_max.t

Alias for Min_max.t.

Sourcetype 'a in_both_abs_axis = 'a In_both_abs_axis.t