package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.Align_selfSource

Controls alignment of an individual child node.

Overrides the parent node's align_items property. For Flexbox it controls alignment in the cross axis. For Grid it controls alignment in the block axis.

See MDN: align-self.

Sourcetype t
Sourceval default : t

default returns Stretch.

Sourceval to_string : t -> string

to_string x converts the alignment value to its CSS string representation.