package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.Flex_wrapSource

Controls whether flex items wrap onto multiple lines.

Defaults to No_wrap.

See CSS Flexbox Specification.

Sourcetype t =
  1. | No_wrap
    (*

    Items stay on a single line

    *)
  2. | Wrap
    (*

    Items wrap to multiple lines as needed

    *)
  3. | Wrap_reverse
    (*

    Items wrap to multiple lines in reverse direction

    *)
Sourceval default : t

default returns No_wrap.

Sourceval to_string : t -> string

to_string t returns the CSS string representation of t.

Sourceval equal : t -> t -> bool

equal a b returns true if a and b are the same wrap mode.

Sourceval compare : t -> t -> int

compare a b compares a and b for use in maps and sets.

Sourceval pp : Format.formatter -> t -> unit

pp fmt t prints t to formatter fmt.