package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Grid_placement.LineSource

Sourceval into_origin_zero_ignoring_named : t Geometry.Line.t -> int -> Grid.Origin_zero_placement.t Geometry.Line.t

into_origin_zero_ignoring_named line explicit_track_count converts both start and end placements to origin-zero coordinates, ignoring named lines.

Applies into_origin_zero_placement_ignoring_named to both the start and end placements.

into_origin_zero line explicit_track_count converts both start and end placements to origin-zero coordinates.

Applies into_origin_zero_placement to both the start and end placements.

Raises Failure if either placement is Named_line or Named_span. Named lines must be resolved before calling this function.

Sourceval is_definite : t Geometry.Line.t -> bool

is_definite line returns true if the line placement is definite.

A line placement is definite if at least one of the start or end positions is a non-zero line index or a named line.