package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.Justify_contentSource

Sets the distribution of space between and around content items.

For Flexbox it controls alignment in the main axis. For Grid it controls alignment in the inline axis.

See MDN: justify-content.

Sourcetype t
Sourceval default : t

default returns Flex_start.

Sourceval to_string : t -> string

to_string x converts the alignment value to its CSS string representation.