package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.Justify_itemsSource

Controls how child nodes are aligned.

Does not apply to Flexbox and will be ignored if specified on a flex container. For Grid it controls alignment in the inline axis.

See MDN: justify-items.

Sourcetype t
Sourceval default : t

default returns Stretch.

Sourceval to_string : t -> string

to_string x converts the alignment value to its CSS string representation.