package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.Justify_selfSource

Controls alignment of an individual child node.

Overrides the parent node's justify_items property. Does not apply to Flexbox and will be ignored if specified on a flex child. For Grid it controls alignment in the inline axis.

See MDN: justify-self.

Sourcetype t
Sourceval default : t

default returns Stretch.

Sourceval to_string : t -> string

to_string x converts the alignment value to its CSS string representation.