package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.PositionSource

CSS position property controlling normal flow participation.

The position property determines whether a node's position is computed by the normal layout flow or positioned relative to its final layout position or parent container.

Sourcetype t =
  1. | Relative
    (*

    The offset is computed relative to the final position given by the layout algorithm. Offsets do not affect the position of any other items.

    *)
  2. | Absolute
    (*

    The node is taken out of the normal flow of layout and positioned relative to its parent.

    *)
Sourceval default : t

default returns Relative.

Sourceval to_string : t -> string

to_string position converts position to its CSS string representation.

Sourceval equal : t -> t -> bool

equal a b returns true if a and b represent the same position mode.

Sourceval compare : t -> t -> int

compare a b returns a comparison result for use in ordered containers.

The ordering is Relative < Absolute.

Sourceval pp : Format.formatter -> t -> unit

pp fmt position prints position to the formatter fmt.