package toffee

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Style.Rect_dimSource

all_px v creates a rectangle with all edges set to v pixels, using Length_percentage.length.

Sourceval all_px_auto : float -> length_percentage_auto Geometry.rect

all_px_auto v creates a rectangle with all edges set to v pixels, for use with auto-capable properties like margin and inset.

Sourceval xy_px : x:float -> y:float -> length_percentage Geometry.rect

xy_px ~x ~y creates a rectangle with horizontal edges set to x pixels and vertical edges set to y pixels.

Sourceval xy_px_auto : x:float -> y:float -> length_percentage_auto Geometry.rect

xy_px_auto ~x ~y is the auto-capable variant of xy_px, suitable for margin and inset.

Sourceval horizontal_px : float -> length_percentage Geometry.rect

horizontal_px v creates a rectangle with left and right set to v pixels and top/bottom set to zero.

Sourceval horizontal_px_auto : float -> length_percentage_auto Geometry.rect

horizontal_px_auto v is the auto-capable variant of horizontal_px.

Sourceval vertical_px : float -> length_percentage Geometry.rect

vertical_px v creates a rectangle with top and bottom set to v pixels and left/right set to zero.

Sourceval vertical_px_auto : float -> length_percentage_auto Geometry.rect

vertical_px_auto v is the auto-capable variant of vertical_px.